import json
import logging
import stripe

from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.http.response import HttpResponse
from django.db.models import TextChoices, Q
from django.utils import timezone
from django.utils.translation import gettext as _
from django.views import View
from django.views.decorators.csrf import csrf_exempt

from vendor.models.choice import InvoiceStatus, PurchaseStatus
from vendor.integrations import StripeIntegration
from vendor.utils import get_site_from_request
from vendor.models import CustomerProfile, Subscription, Invoice
from vendor.processors import StripeProcessor
from vendor.config import SupportedPaymentProcessor


logger = logging.getLogger(__name__)



# TODO: Need to add more validation to function example:
# The lowest number can only 50 which transaltes to $0.50
# Probably should also added it to the processor as a static function
def convert_integer_to_float(number):
    number_string = str(number)

    return float(f"{number_string[:-2]}.{number_string[-2:]}")


class StripeEvents(TextChoices):
    INVOICE_PAID = 'invoice.paid', _('Invoice Paid')
    INVOICE_PAYMENT_FAILED = 'invoice.payment_failed', _('Invoice Payment Failed')
    INOVICE_PAYMENT_SUCCEEDED = 'invoice.payment_succeeded', _('Invoice Payment Succeeded')
    PAYMENT_INTENT_SUCCEDED = 'payment_intent.succeeded', _("Payment Succeeded")
    CHARGE_SUCCEEDED = 'charge.succeeded', _('Charge Succeeded')
    SOURCE_EXPIRED = 'customer.source.expired', _('Source Expired')



class StripeBaseAPI(View):

    def __init__(self, **kwargs):
        self.stripe_event = None  # Variable used to store the webhooks event data.

        super().__init__(**kwargs)

    @csrf_exempt
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)

    def is_valid_post(self, site):
        try:
            credentials = StripeIntegration(site)

            if credentials.instance and credentials.instance.private_key:
                self.event = stripe.Event.construct_from(json.loads(self.request.body), credentials.instance.private_key)
            elif settings.STRIPE_PUBLIC_KEY:
                self.event = stripe.Event.construct_from(json.loads(self.request.body), settings.STRIPE_PUBLIC_KEY)
            else:
                return False
            
        except ValueError as exce:
            return False
        except Exception as exce:
            return False

        return True
    
    def is_incoming_event_correct_and_recurring(self, event, desired_event):
        if event.type != desired_event:
            return False

        # This check is recuired to make sure that the event is related to a subscription.
        if self.event.data.object.billing_reason != 'subscription_cycle':
            return False

        return True
    
    def is_incoming_event_correct(self, event, desired_event):
        if event.type != desired_event:
            return False

        return True


class StripeSubscriptionInvoicePaid(StripeBaseAPI):

    def post(self, request, *args, **kwargs):
        site = get_site_from_request(self.request)

        if not self.is_valid_post(site):
            logger.error(f"StripeSubscriptionInvoicePaid error: invalid post")
            return HttpResponse(status=400, content="StripeSubscriptionInvoicePaid invalid post")

        if not self.is_incoming_event_correct_and_recurring(self.event, StripeEvents.INVOICE_PAID):
            logger.error(f"StripeSubscriptionInvoicePaid error: invalid event: {self.event}")
            return HttpResponse(status=200, content=f"StripeSubscriptionInvoicePaid error: invalid event: {self.event}")

        stripe_invoice = self.event.data.object
        paid_date = timezone.datetime.fromtimestamp(stripe_invoice.status_transitions.paid_at, tz=timezone.utc)
        amount_paid = convert_integer_to_float(stripe_invoice.total)

        # TODO nice to move this try/except blocks inside a generic function in the StripeBaseAPI class
        try:
            customer_profile = CustomerProfile.objects.get(site=site, user__email=stripe_invoice.customer_email)
        except ObjectDoesNotExist:
            logger.error(f"StripeSubscriptionInvoicePaid error: email: {stripe_invoice.customer_email} does not exist")
            return HttpResponse(status=400, content=f"StripeSubscriptionInvoicePaid error: email: {stripe_invoice.customer_email} does not exist")

        if 'stripe_id' not in customer_profile.meta:
            customer_profile.meta['stripe_id'] = stripe_invoice.customer
            customer_profile.save()

        try:
            subscription = Subscription.objects.get(gateway_id=stripe_invoice.subscription, profile=customer_profile)
        except ObjectDoesNotExist:
            logger.error(f"StripeSubscriptionInvoicePaid customer: {customer_profile} does not have a subscription with stripe_id: {stripe_invoice.subscription}")
            return HttpResponse(status=400, content=f"StripeSubscriptionInvoicePaid customer: {customer_profile} does not have a subscription with stripe_id: {stripe_invoice.subscription}")

        offer = subscription.get_offer()

        if not offer:
            logger.error(f"StripeSubscriptionInvoicePaid {subscription} has no offer attached")
            return HttpResponse(status=400, content=f"StripeSubscriptionInvoicePaid {subscription} has no offer attached")

        invoice = Invoice.objects.create(
            profile=customer_profile,
            site=site,
            ordered_date=paid_date,
            status=InvoiceStatus.COMPLETE
        )
        invoice.add_offer(offer)
        invoice.total = amount_paid
        invoice.vendor_notes = {'stripe_id': stripe_invoice.stripe_id}
        invoice.save()

        processor = StripeProcessor(site, invoice)
        processor.renew_subscription(subscription, stripe_invoice.charge, PurchaseStatus.SETTLED, payment_success=True, submitted_date=paid_date)

        return HttpResponse(status=200)


class StripeSubscriptionPaymentFailed(StripeBaseAPI):

    def post(self, request, *args, **kwargs):
        site = get_site_from_request(request)

        if not self.is_valid_post(site):
            logger.error("StripeSubscriptionPaymentFailed error: invalid post")
            return HttpResponse(status=400, content="StripeSubscriptionPaymentFailed error: invalid post")

        if not self.is_incoming_event_correct_and_recurring(self.event, StripeEvents.INVOICE_PAYMENT_FAILED):
            logger.error(f"StripeSubscriptionPaymentFailed error: invalid event: {self.event}")
            return HttpResponse(status=200, content=f"StripeSubscriptionPaymentFailed error: invalid event: {self.event}")

        stripe_invoice = self.event.data.object
        paid_date = timezone.datetime.fromtimestamp(stripe_invoice.status_transitions.paid_at, tz=timezone.utc)

        # TODO nice to move this try/except blocks inside a generic function in the StripeBaseAPI class
        try:
            customer_profile = CustomerProfile.objects.get(site=site, user__email=stripe_invoice.customer_email)
        except ObjectDoesNotExist:
            logger.error(f"StripeSubscriptionPaymentFailed error: email: {stripe_invoice.customer_email} does not exist")
            return HttpResponse(status=400, content=f"StripeSubscriptionPaymentFailed error: email: {stripe_invoice.customer_email} does not exist")

        if 'stripe_id' not in customer_profile.meta:
            customer_profile.meta['stripe_id'] = stripe_invoice.customer
            customer_profile.save()

        try:
            subscription = Subscription.objects.get(gateway_id=stripe_invoice.subscription, profile=customer_profile)
        except ObjectDoesNotExist:
            logger.error(f"StripeSubscriptionPaymentFailed customer: {customer_profile} does not have a subscription with stripe_id: {stripe_invoice.subscription}")
            return HttpResponse(status=400, content=f"StripeSubscriptionPaymentFailed customer: {customer_profile} does not have a subscription with stripe_id: {stripe_invoice.subscription}")

        offer = subscription.get_offer()

        if not offer:
            logger.error(f"StripeSubscriptionPaymentFailed {subscription} has no offer attached")
            return HttpResponse(status=400, content=f"StripeSubscriptionPaymentFailed {subscription} has no offer attached")

        invoice = Invoice.objects.create(
            profile=subscription.profile,
            site=subscription.profile.site,
            ordered_date=paid_date,
            status=InvoiceStatus.COMPLETE
        )

        invoice.add_offer(offer)
        invoice.vendor_notes = {}
        invoice.vendor_notes['stripe_id'] = stripe_invoice.stripe_id
        invoice.save()

        processor = StripeProcessor(site, invoice)
        processor.subscription_payment_failed(subscription, stripe_invoice.charge)

        return HttpResponse(status=200)


class StripeInvoicePaid(StripeBaseAPI):

    def post(self, request, *args, **kwargs):
        site = get_site_from_request(self.request)

        if not self.is_valid_post(site):
            logger.error("StripeInvoicePaid error: invalid post")
            return HttpResponse(status=400, content="StripeInvoicePaid error: invalid post")

        if not self.is_incoming_event_correct(self.event, StripeEvents.INOVICE_PAYMENT_SUCCEEDED):
            logger.error(f"StripeInvoicePaid error: invalid event {self.event}")
            return HttpResponse(status=200, content=f"StripeInvoicePaid error: invalid event {self.event}")

        stripe_invoice = self.event.data.object

        try:
            customer_profile = CustomerProfile.objects.get(site=site, user__email=stripe_invoice.customer_email)
        except ObjectDoesNotExist:
            logger.error(f"StripeInvoicePaid error: email: {stripe_invoice.customer_email} does not exist")
            return HttpResponse(status=400, content=f"StripeInvoicePaid error: email: {stripe_invoice.customer_email} does not exist")

        if 'stripe_id' not in customer_profile.meta:
            customer_profile.meta['stripe_id'] = stripe_invoice.customer
            customer_profile.save()

        try:
            subscription = Subscription.objects.get(gateway_id=stripe_invoice.subscription, profile=customer_profile)
        except ObjectDoesNotExist:
            logger.error(f"StripeInvoicePaid customer: {customer_profile} does not have a subscription with stripe_id: {stripe_invoice.subscription}")
            return HttpResponse(status=400, content=f"StripeInvoicePaid customer: {customer_profile} does not have a subscription with stripe_id: {stripe_invoice.subscription}")

        payment = subscription.payments.filter(Q(transaction="") | Q(transaction=None), status=PurchaseStatus.QUEUED).first()
        receipt = subscription.receipts.filter(Q(transaction="") | Q(transaction=None)).first()

        if not payment and not receipt:
            logger.warning(f"There are no payments to update for subscription: {subscription}. Stripe Invoice: {stripe_invoice.id}")
            return HttpResponse(status=200, content=f"There are no payments to update for subscription: {subscription}. Stripe Invoice: {stripe_invoice.id}")
        
        if payment:
            payment.transaction = stripe_invoice.charge
            payment.status = PurchaseStatus.SETTLED
            payment.save()

        if receipt:
            receipt.transaction = stripe_invoice.charge
            receipt.save()

        return HttpResponse(status=200)


class StripeCardExpiring(StripeBaseAPI):

    def post(self, request, *args, **kwargs):
        site = get_site_from_request(self.request)

        if not self.is_valid_post(site):
            logger.error("StripeCardExpiring error: invalid event post")
            return HttpResponse(status=400, content="StripeCardExpiring error: invalid event post")

        if not self.is_incoming_event_correct(self.event, StripeEvents.SOURCE_EXPIRED):
            logger.error(f"StripeCardExpiring error: invalid event {self.event}")
            return HttpResponse(status=200, content=f"StripeCardExpiring error: invalid event {self.event}")

        stripe_card = self.event.data.object

        stripe_customer_id = stripe_card['customer']
        if stripe_customer_id:
            try:
                customer_profile = CustomerProfile.objects.get(meta__stripe_id=stripe_customer_id)
            except ObjectDoesNotExist:
                logger.error(f"StripeCardExpiring: stripe id {stripe_customer_id} does not exist for customer in vendor")
                return HttpResponse(status=200, content=f"StripeCardExpiring: stripe id {stripe_customer_id} does not exist for customer in vendor")

            email = customer_profile.user.email
            logger.info(f'StripeCardExpiring: sending customer_source_expiring signal for site {site} and email {email}')
            
            processor = StripeProcessor(site)
            processor.customer_card_expired(site, email)
            
        return HttpResponse(status=200)


class StripeSyncObjects(View):

    def get(self, request, *args, **kwargs):
        site = get_site_from_request(request)
        logger.info(f'StripeSyncObjects: started for site: {site}')

        processor = StripeProcessor(site)
        processor.sync_stripe_vendor_objects(site)

        logger.info(f'StripeSyncObjects: finished for site: {site}')
        return HttpResponse(status=200)




# Generated by Django 3.1.3 on 2020-11-23 18:48

from django.db import migrations, models
import vendor.models.offer


class Migration(migrations.Migration):

    dependencies = [
        ('vendor', '0011_auto_20201120_1750'),
    ]

    operations = [
        migrations.AlterField(
            model_name='offer',
            name='offer_description',
            field=models.TextField(blank=True, default=None, help_text='You can enter a list of descriptions. Note: if you inputs something here the product description will not show up.', null=True, verbose_name='Offer Description'),
        ),
        migrations.AlterField(
            model_name='offer',
            name='term_details',
            field=models.JSONField(blank=True, default=vendor.models.offer.offer_term_details_default, help_text='term_units: 10/20(Day/Month), trial_occurrences: 1(defualt)', null=True, verbose_name='Term Details'),
        ),
    ]



from django.contrib.auth.mixins import LoginRequiredMixin
from django.contrib.sites.models import Site
from django.views.generic.list import ListView
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views.generic.edit import FormView, UpdateView
from django.shortcuts import redirect, render

from vendor.config import PaymentProcessorSiteConfig, PaymentProcessorForm,\
    StripeConnectAccountConfig, StripeConnectAccountForm,\
    VendorSiteCommissionForm, VendorSiteCommissionConfig
from siteconfigs.models import SiteConfigModel


class PaymentProcessorSiteConfigsListView(LoginRequiredMixin, ListView):
    template_name = 'vendor/manage/config_list.html'
    model = SiteConfigModel

    def get_queryset(self):
        payment_processor = PaymentProcessorSiteConfig()

        return SiteConfigModel.objects.filter(key=payment_processor.key)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        processor_config = PaymentProcessorSiteConfig()

        context['title'] = _('Payment Processors Configured')
        context['config_key'] = processor_config.key
        context['new_url'] = reverse('vendor_admin:manager-config-processor-create')

        return context


class PaymentProcessorCreateConfigView(LoginRequiredMixin, FormView):
    template_name = 'vendor/manage/config.html'
    form_class = PaymentProcessorForm

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(**kwargs)
        processor_config = PaymentProcessorSiteConfig()
        form = self.get_form()

        existing_stripe_configs = SiteConfigModel.objects.filter(key=processor_config.key)
        sites = Site.objects.exclude(pk__in=[config.site.pk for config in existing_stripe_configs])
        form.fields['site'].queryset = sites

        context['title'] = _("Payment Processor Config")
        context['form'] = form
        context['cancel_url'] = reverse('vendor_admin:manager-config-processor-list')

        return context

    def form_valid(self, form):
        site = form.cleaned_data['site']

        processor_config = PaymentProcessorSiteConfig(site)
        processor_config.save(form.cleaned_data["payment_processor"], "payment_processor")
        
        return redirect('vendor_admin:manager-config-processor-list')


class PaymentProcessorUpdateConfigView(LoginRequiredMixin, UpdateView):
    template_name = 'vendor/manage/config.html'
    model = SiteConfigModel
    form_class = PaymentProcessorForm

    def get_form(self):
        return PaymentProcessorForm()

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        config_model = self.get_object()
        processor_config = PaymentProcessorSiteConfig(config_model.site)

        context['title'] = _("Edit Payment Processor Config")
        context['form'] = PaymentProcessorForm(initial={'site': config_model.site, 'payment_processor': processor_config.get_key_value('payment_processor')})
        context['form'].fields['site'].disabled = True
        context['cancel_url'] = reverse('vendor_admin:manager-config-processor-list')

        return context


    def post(self, request, *args, **kwargs):
        # Get the SiteConfigModel instance passed through the url
        self.object = self.get_object()

        form = self.get_form_class()(request.POST)
        # Make this site field required False since you can only update the account_number
        form.fields['site'].required = False

        if not form.is_valid():
            context = super().get_context_data(**kwargs)
            context['form'] = form
            return render(request, self.template_name, context)
        
        processor_config = PaymentProcessorSiteConfig(self.object.site)
        processor_config.save(form.cleaned_data['payment_processor'], "payment_processor")

        return redirect("vendor_admin:manager-config-processor-list")


class StripeConnectAccountCreateConfigView(LoginRequiredMixin, FormView):
    template_name = 'vendor/manage/config.html'
    form_class = StripeConnectAccountForm

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        stripe_config = StripeConnectAccountConfig()
        form = self.get_form()

        existing_stripe_configs = SiteConfigModel.objects.filter(key=stripe_config.key)
        sites = Site.objects.exclude(pk__in=[config.site.pk for config in existing_stripe_configs])
        form.fields['site'].queryset = sites

        context['title'] = _("Stripe Connect Account")
        context['form'] = form
        context['cancel_url'] = reverse('vendor_admin:manager-config-stripe-connect-list')

        return context

    def form_valid(self, form):
        site = form.cleaned_data['site']
        
        stripe_connect = StripeConnectAccountConfig(site)
        stripe_connect.save(form.cleaned_data['account_number'], "stripe_connect_account")

        return redirect("vendor_admin:manager-config-stripe-connect-list")


class StripeConnectAccountConfigListView(LoginRequiredMixin, ListView):
    template_name = 'vendor/manage/config_list.html'
    model = SiteConfigModel

    def get_queryset(self):
        stripe_config = StripeConnectAccountConfig()
        
        return self.model.objects.filter(key=stripe_config.key)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        stripe_config = StripeConnectAccountConfig()

        context['title'] = _('Stripe Connect Accounts')
        context['config_key'] = stripe_config.key
        context['new_url'] = reverse('vendor_admin:manager-config-stripe-connect-create')

        return context


class StripeConnectAccountUpdateConfigView(LoginRequiredMixin, UpdateView):
    template_name = 'vendor/manage/config.html'
    model = SiteConfigModel
    form_class = StripeConnectAccountForm

    def get_form(self):
        return StripeConnectAccountForm()

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        config_model = self.get_object()
        stripe_config = StripeConnectAccountConfig(config_model.site)

        context['title'] = _("Stripe Connect Account")
        context['form'] = StripeConnectAccountForm(initial={'site': config_model.site, 'account_number': stripe_config.get_key_value('stripe_connect_account')})
        context['form'].fields['site'].disabled = True
        context['cancel_url'] = reverse('vendor_admin:manager-config-stripe-connect-list')

        return context


    def post(self, request, *args, **kwargs):
        # Get the SiteConfigModel instance passed through the url
        self.object = self.get_object()

        form = self.get_form_class()(request.POST)
        # Make this site field required False since you can only update the account_number
        form.fields['site'].required = False

        if not form.is_valid():
            context = super().get_context_data(**kwargs)
            context['form'] = form
            return render(request, self.template_name, context)
        
        stripe_connect = StripeConnectAccountConfig(self.object.site)
        stripe_connect.save(form.cleaned_data['account_number'], "stripe_connect_account")

        return redirect("vendor_admin:manager-config-stripe-connect-list")


class VendorSiteCommissionCreateConfigView(LoginRequiredMixin, FormView):
    template_name = 'vendor/manage/config.html'
    form_class = VendorSiteCommissionForm

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        commission_config = VendorSiteCommissionConfig()
        form = self.get_form()

        existing_commission_configs = SiteConfigModel.objects.filter(key=commission_config.key)
        sites = Site.objects.exclude(pk__in=[config.site.pk for config in existing_commission_configs])
        form.fields['site'].queryset = sites

        context['title'] = _("Vendor Site Commission")
        context['form'] = form
        context['cancel_url'] = reverse('vendor_admin:manager-config-commission-list')

        return context

    def form_valid(self, form):
        site = form.cleaned_data['site']
        
        commission = VendorSiteCommissionConfig(site)
        commission.save(form.cleaned_data['commission'], "commission")

        return redirect("vendor_admin:manager-config-commission-list")


class VendorSiteCommissionConfigListView(LoginRequiredMixin, ListView):
    template_name = 'vendor/manage/config_list.html'
    model = SiteConfigModel

    def get_queryset(self):
        commission_config = VendorSiteCommissionConfig()
        
        return self.model.objects.filter(key=commission_config.key)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        commission_config = VendorSiteCommissionConfig()

        context['title'] = _('Vendor Commissions')
        context['config_key'] = commission_config.key
        context['new_url'] = reverse('vendor_admin:manager-config-commission-create')

        return context


class VendorSiteCommissionUpdateConfigView(LoginRequiredMixin, UpdateView):
    template_name = 'vendor/manage/config.html'
    model = SiteConfigModel
    form_class = VendorSiteCommissionForm

    def get_form(self):
        return VendorSiteCommissionForm()

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        config_model = self.get_object()
        commission_config = VendorSiteCommissionConfig(config_model.site)

        context['title'] = _("Vendor Commission")
        context['form'] = VendorSiteCommissionForm(initial={'site': config_model.site, 'commission': commission_config.get_key_value('commission')})
        context['form'].fields['site'].disabled = True
        context['cancel_url'] = reverse('vendor_admin:manager-config-commission-list')

        return context

    def post(self, request, *args, **kwargs):
        # Get the SiteConfigModel instance passed through the url
        self.object = self.get_object()

        form = self.get_form_class()(request.POST)
        # Make this site field required False since you can only update the commission
        form.fields['site'].required = False

        if not form.is_valid():
            context = super().get_context_data(**kwargs)
            context['form'] = form
            return render(request, self.template_name, context)
        
        stripe_connect = VendorSiteCommissionConfig(self.object.site)
        stripe_connect.save(form.cleaned_data['commission'], "commission")

        return redirect("vendor_admin:manager-config-commission-list")



# Generated by Django 3.1.3 on 2021-10-04 19:45

from django.db import migrations


def change_list_to_dict_vendor_notes(apps, schema_editor):
    InvoiceModel = apps.get_model('vendor', 'Invoice')
    for invoice in InvoiceModel.objects.all():
        if isinstance(invoice.vendor_notes, list):
            temp_list = invoice.vendor_notes
            invoice.vendor_notes = {}
            invoice.vendor_notes['bugfix_migration_0027'] = temp_list
            invoice.save()


class Migration(migrations.Migration):

    dependencies = [
        ('vendor', '0026_auto_20210914_1209'),
    ]

    operations = [
        migrations.RunPython(change_list_to_dict_vendor_notes, reverse_code=migrations.RunPython.noop),
    ]



# Generated by Django 3.2.13 on 2022-06-01 18:58

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('vendor', '0031_pre_invoice_status_change'),
    ]

    operations = [
        migrations.AlterField(
            model_name='invoice',
            name='status',
            field=models.IntegerField(choices=[(0, 'Cart'), (10, 'Checkout'), (20, 'Complete')], default=0, verbose_name='Status'),
        ),
    ]



# Generated by Django 3.2.16 on 2023-01-18 18:06

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('vendor', '0040_auto_20221020_1617'),
    ]

    operations = [
        migrations.AddField(
            model_name='offer',
            name='billing_start_date',
            field=models.DateTimeField(blank=True, default=None, help_text='The start date to begin billing', null=True, verbose_name='Billing Start Date'),
        ),
    ]



# App Settings
from django.conf import settings
from django import forms
from django.db.models import TextChoices
from django.contrib.sites.models import Site
from django.utils.translation import gettext_lazy as _

from siteconfigs.config import SiteConfigBaseClass


class SiteSelectForm(forms.Form):
    site = forms.ModelChoiceField(queryset=Site.objects.all())


class SupportedPaymentProcessor(TextChoices):
    PROMO_CODE_BASE = ("base.PaymentProcessorBase", _("Default Processor"))
    AUTHORIZE_NET = ("authorizenet.AuthorizeNetProcessor", _("Authorize.Net"))
    STRIPE = ("stripe_processor.StripeProcessor", _("Stripe"))


class PaymentProcessorForm(SiteSelectForm):
    payment_processor = forms.CharField(label=_("Payment Processor"), widget=forms.Select(choices=SupportedPaymentProcessor.choices))


class PaymentProcessorSiteConfig(SiteConfigBaseClass):
    label = _("Payment Processor")
    default = {"payment_processor": "base.PaymentProcessorBase"}
    form_class = PaymentProcessorForm
    key = ""
    instance = None

    def __init__(self, site=None):
        
        if site is None:
            site = Site.objects.get_current()

        self.key = ".".join([__name__, __class__.__name__])
        super().__init__(site, self.key)
        
        if not self.instance:
            self.default['payment_processor'] = VENDOR_PAYMENT_PROCESSOR

    def get_form(self):
        return self.form_class(initial=self.get_initials())

    def get_initials(self):
        initial = super().get_initials()
        initial['site'] = (self.site.pk, self.site.domain)
        
        if self.instance:
            initial['payment_processor'] = [choice for choice in SupportedPaymentProcessor.choices if choice[0] == self.instance.value['payment_processor']][0]
        else:
            initial['payment_processor'] = SupportedPaymentProcessor.choices[0]
        
        return initial

    def get_selected_processor(self):
        if self.instance:
            return [choice for choice in SupportedPaymentProcessor.choices if choice[0] == self.instance.value['payment_processor']][0]

        return SupportedPaymentProcessor.choices[0]  # Return Default Processors


class StripeConnectAccountForm(SiteSelectForm):
    account_number = forms.CharField(max_length=120)


class StripeConnectAccountConfig(SiteConfigBaseClass):
    label = _("Stripe Connect Account")
    default = {'stripe_connect_account': None}
    form_class = StripeConnectAccountForm
    key = ''
    instance = None

    def __init__(self, site=None):
        
        if site is None:
            site = Site.objects.get_current()

        self.key = ".".join([__name__, __class__.__name__])
        super().__init__(site, self.key)
        

class VendorSiteCommissionForm(SiteSelectForm):
    commission = forms.IntegerField(min_value=0, max_value=100)


class VendorSiteCommissionConfig(SiteConfigBaseClass):
    label = _("Vendor Site Commission")
    default = {'commission': None}
    form_class = VendorSiteCommissionForm
    key = ''
    instance = None

    def __init__(self, site=None):
        
        if site is None:
            site = Site.objects.get_current()

        self.key = ".".join([__name__, __class__.__name__])
        super().__init__(site, self.key)


VENDOR_PRODUCT_MODEL = getattr(settings, "VENDOR_PRODUCT_MODEL", "vendor.Product")

VENDOR_PAYMENT_PROCESSOR = getattr(settings, "VENDOR_PAYMENT_PROCESSOR", "dummy.DummyProcessor")

DEFAULT_CURRENCY = getattr(settings, "DEFAULT_CURRENCY", "usd")

AVAILABLE_CURRENCIES = getattr(settings, "AVAILABLE_CURRENCIES", {'usd': _('USD Dollars')})

VENDOR_STATE = getattr(settings, "VENDOR_STATE", "DEBUG")

# Encryption settings
VENDOR_DATA_ENCODER = getattr(settings, "VENDOR_DATA_ENCODER", "vendor.encrypt.cleartext")





from django.urls import path

from vendor.views import vendor_admin as admin_views
from vendor.views import config as config_views
from vendor.views import integration as integrations_views
from vendor.views import report as report_views

app_name = "vendor_admin"

urlpatterns = [
    path('', admin_views.AdminDashboardView.as_view(), name="manager-dashboard"),
    path('products/', admin_views.AdminProductListView.as_view(), name="manager-product-list"),
    path('product/<uuid:uuid>/', admin_views.AdminProductUpdateView.as_view(), name="manager-product-update"),
    path('product/', admin_views.AdminProductCreateView.as_view(), name="manager-product-create"),
    path('offers/', admin_views.AdminOfferListView.as_view(), name="manager-offer-list"),
    path('offer/<uuid:uuid>/', admin_views.AdminOfferUpdateView.as_view(), name="manager-offer-update"),
    path('offer/', admin_views.AdminOfferCreateView.as_view(), name="manager-offer-create"),
    path('orders/', admin_views.AdminInvoiceListView.as_view(), name="manager-order-list"),
    path('order/<uuid:uuid>/', admin_views.AdminInvoiceDetailView.as_view(), name="manager-order-detail"),
    path('subscriptions/', admin_views.AdminSubscriptionListView.as_view(), name="manager-subscriptions"),
    path('subscription/<uuid:uuid>', admin_views.AdminSubscriptionDetailView.as_view(), name="manager-subscription"),
    path('subscription/<uuid:uuid_subscription>/add/payment/<uuid:uuid_profile>/', admin_views.AdminSubscriptionAddPaymentView.as_view(), name="manager-subscription-add-payment"),
    path('subscription/create/', admin_views.AdminSubscriptionCreateView.as_view(), name="manager-subscription-create"),
    path('profiles/', admin_views.AdminProfileListView.as_view(), name="manager-profiles"),
    path('profile/<uuid:uuid>', admin_views.AdminProfileDetailView.as_view(), name="manager-profile"),
    path('product/<uuid:uuid>/renew', admin_views.AdminManualSubscriptionRenewal.as_view(), name="manager-product-renew"),
    path('payments/no/receipt/', admin_views.PaymentWithNoReceiptListView.as_view(), name="manager-payment-no-receipt"),
    path('payments/no/orderitems/', admin_views.PaymentWithNoOrderItemsListView.as_view(), name="manager-payment-no-receipt"),

    # Reports
    path('reports/receipts/download/', report_views.ReceiptListCSV.as_view(), name="manager-receipt-download"),
    path('reports/invoices/download/', report_views.InvoiceListCSV.as_view(), name="manager-invoice-download"),

    # Integrations
    path("authorizenet/integration/", integrations_views.AuthorizeNetIntegrationView.as_view(), name="authorizenet-integration"),
    path("stripe/integration/", integrations_views.StripeIntegrationView.as_view(), name="stripe-integration"),

    # Configs
    path('config/stripe/connect/list/', config_views.StripeConnectAccountConfigListView.as_view(), name="manager-config-stripe-connect-list"),
    path('config/stripe/connect/create/', config_views.StripeConnectAccountCreateConfigView.as_view(), name="manager-config-stripe-connect-create"),
    path('config/stripe/connect/update/<int:pk>/', config_views.StripeConnectAccountUpdateConfigView.as_view(), name="manager-config-stripe-connect-update"),
    path('config/commission/list/', config_views.VendorSiteCommissionConfigListView.as_view(), name="manager-config-commission-list"),
    path('config/commission/create/', config_views.VendorSiteCommissionCreateConfigView.as_view(), name="manager-config-commission-create"),
    path('config/commission/update/<int:pk>/', config_views.VendorSiteCommissionUpdateConfigView.as_view(), name="manager-config-commission-update"),
    # path('config/vendor/commission/', config_views.InvoiceListCSV.as_view(), name="manager-invoice-download"),
    path("config/processor/list/", config_views.PaymentProcessorSiteConfigsListView.as_view(), name="manager-config-processor-list"),
    path("config/processor/create/", config_views.PaymentProcessorCreateConfigView.as_view(), name="manager-config-processor-create"),
    path("config/processor/update/<int:pk>", config_views.PaymentProcessorUpdateConfigView.as_view(), name="manager-config-processor-update"),
    
]



import importlib.metadata

try:
    __version__ = importlib.metadata.version("vendor")
    print("DIR NAME")
    print(os.path.dirname(__file__))
except:     # This is to support installing the package locally via "pip install -e ." where a metadata files is not yet been created.
    import toml, os
    module_dir = os.path.dirname(__file__)
    cfg = toml.load(module_dir + "/../../pyproject.toml")
    __version__ = cfg['project']['version']

VERSION = __version__



from django.core.exceptions import ObjectDoesNotExist
from integrations.models import Credential
from vendor.forms import AuthorizeNetIntegrationForm, StripeIntegrationForm


class BaseIntegration(object):
    NAME = None

    site = None
    form_class = None

    def __init__(self, site):
        if not self.NAME:
            raise Exception('NAME is required for all BaseIntegration instances')

        self.site = site
        self.instance = self.get_instance()

    def get_instance(self):
        try:
            return Credential.objects.get(name=self.NAME, site=self.site)
        except ObjectDoesNotExist:
            return None

    def save(self, data):
        if not self.instance:
            form = self.form_class(data)
        else:
            form = self.form_class(data, instance=self.instance)

        integration_form = form.save(commit=False)
        integration_form.name = self.NAME
        integration_form.site = self.site
        integration_form.save()


class AuthorizeNetIntegration(BaseIntegration):
    NAME = "AuthorizeNet Integration"

    site = None
    form_class = AuthorizeNetIntegrationForm


class StripeIntegration(BaseIntegration):
    NAME = "Stripe Integration"

    site = None
    form_class = StripeIntegrationForm



import logging

from django.contrib import messages
from django.contrib.auth.mixins import LoginRequiredMixin
from django.core.exceptions import ObjectDoesNotExist
from django.shortcuts import Http404, redirect, render
from django.urls import reverse_lazy
from django.utils.translation import gettext as _
from django.views.generic import FormView, TemplateView
from django.views.generic.detail import DetailView
from django.views.generic.edit import UpdateView
from django.views.generic.list import ListView

from vendor.forms import (AccountInformationForm, AddressForm,
                          BillingAddressForm, CreditCardForm)
from vendor.models import (Address, Invoice, Offer, OrderItem, Receipt,
                           Subscription)
from vendor.models.choice import InvoiceStatus, TermType
from vendor.processors import get_site_payment_processor
from vendor.utils import (clear_session_purchase_data,
                          get_or_create_session_cart, get_site_from_request)

logger = logging.getLogger(__name__)


# TODO: Need to remove the login required
def get_purchase_invoice(user, site):
    """
    Return an invoice that is in checkout or cart state or a newly create invoice in cart state.
    """
    profile, created = user.customer_profile.get_or_create(site=site)
    return profile.get_cart_or_checkout_cart()


def check_offer_items_or_redirect(invoice, request):

    if invoice.order_items.count() < 1:
        messages.info(request, _("Please add to your cart"))
        redirect('vendor:cart')


class VendorHomeView(TemplateView):
    template_name = 'vendor/index.html'


class CartView(TemplateView):
    '''
    View items in the cart
    '''
    template_name = 'vendor/invoice_detail.html'

    def get(self, request, *args, **kwargs):
        context = super().get_context_data(**kwargs)
        if request.user.is_anonymous:
            session_cart = get_or_create_session_cart(request.session)

            context['invoice'] = {}
            if len(session_cart):
                context['order_items'] = [ OrderItem(offer=Offer.objects.get(pk=offer), quantity=session_cart[offer]['quantity']) for offer in session_cart.keys() ]
            else:
                context['order_items'] = []
            context['invoice']['subtotal'] = sum([item.total for item in context['order_items'] ])
            context['invoice']['shipping'] = 0
            context['invoice']['tax'] = 0
            context['invoice']['total'] = context['invoice']['subtotal']

            return render(request, self.template_name, context)

        profile, created = self.request.user.customer_profile.get_or_create(site=get_site_from_request(request))
        cart = profile.get_cart_or_checkout_cart()
        context['invoice'] = cart
        context['order_items'] = [ order_item for order_item in cart.order_items.all() ]
        return render(request, self.template_name, context)


class AccountInformationView(LoginRequiredMixin, TemplateView):
    template_name = 'vendor/checkout.html'

    def get_form_class(self):
        return AccountInformationForm

    def get(self, request, *args, **kwargs):
        context = super().get_context_data(**kwargs)
        clear_session_purchase_data(request)

        invoice = get_purchase_invoice(request.user, get_site_from_request(request))
        
        if not invoice.order_items.count():
            return redirect('vendor:cart')

        invoice.status = InvoiceStatus.CHECKOUT
        invoice.save()

        existing_account_address = Address.objects.filter(profile__user=request.user, profile__site=get_site_from_request(request)).last()

        form_class = self.get_form_class()
        if existing_account_address:
            # TODO: In future the user will be able to select from multiple saved address
            form = form_class(initial={'email': request.user.email}, instance=existing_account_address)
        else:
            form = form_class(initial={'first_name': request.user.first_name, 'last_name': request.user.last_name, 'email': request.user.email})

        context['form'] = form
        context['invoice'] = invoice

        return render(request, self.template_name, context)

    def post(self, request, *args, **kwargs):
        form = self.get_form_class()(request.POST)
        invoice = get_purchase_invoice(request.user, get_site_from_request(request))

        if not form.is_valid():
            return render(request, self.template_name, {'form': form, 'invoice': invoice})

        shipping_address = form.save(commit=False)

        if not invoice.order_items.count() or invoice.status == InvoiceStatus.CART:
            messages.info(request, _("Cart changed while in checkout process"))
            return redirect('vendor:cart')

        invoice.status = InvoiceStatus.CHECKOUT
        invoice.customer_notes = {'remittance_email': form.cleaned_data['email']}
        # TODO: Need to add a drop down to select existing address
        shipping_address, created = invoice.profile.get_or_create_address(shipping_address)

        if created:
            shipping_address.profile = invoice.profile
            shipping_address.save()
        invoice.shipping_address = shipping_address
        invoice.save()

        return redirect('vendor:checkout-payment')


class PaymentView(LoginRequiredMixin, TemplateView):
    '''
    Review items and submit Payment
    '''
    template_name = "vendor/checkout.html"

    def get(self, request, *args, **kwargs):
        invoice = get_purchase_invoice(request.user, get_site_from_request(request))

        if not invoice.order_items.count():
            return redirect('vendor:cart')

        context = super().get_context_data()

        processor = get_site_payment_processor(invoice.site)(invoice.site, invoice)

        context = processor.get_checkout_context(context=context)

        return render(request, self.template_name, context)

    def post(self, request, *args, **kwargs):
        context = super().get_context_data(**kwargs)
        invoice = get_purchase_invoice(request.user, get_site_from_request(request))

        if not invoice.order_items.count() or invoice.status == InvoiceStatus.CART:
            messages.info(request, _("Cart changed while in checkout process"))
            return redirect('vendor:cart')

        credit_card_form = CreditCardForm(request.POST)
        
        if request.POST.get('billing-same_as_shipping') == 'on':
            billing_address_form = BillingAddressForm(instance=invoice.shipping_address)
            billing_address_form.data = {f'billing-{key}': value for key, value in billing_address_form.initial.items()}
            billing_address_form.is_bound = True
        else:
            billing_address_form = BillingAddressForm(request.POST)

        if not (billing_address_form.is_valid() and credit_card_form.is_valid()):
            processor = get_site_payment_processor(invoice.site)(invoice.site, invoice)
            context['billing_address_form'] = billing_address_form
            context['credit_card_form'] = credit_card_form
            return render(request, self.template_name, processor.get_checkout_context(context=context))
        else:
            billing_address_form.full_clean()
            credit_card_form.full_clean()
            request.session['billing_address_form'] = {f'billing-{key}': value for key, value in billing_address_form.cleaned_data.items()}
            request.session['credit_card_form'] = credit_card_form.cleaned_data
            return redirect('vendor:checkout-review')


class ReviewCheckoutView(LoginRequiredMixin, TemplateView):
    template_name = 'vendor/checkout.html'

    def get(self, request, *args, **kwargs):
        invoice = get_purchase_invoice(request.user, get_site_from_request(request))
        if not invoice.order_items.count():
            return redirect('vendor:cart')

        context = super().get_context_data()

        processor = get_site_payment_processor(invoice.site)(invoice.site, invoice)
        if 'billing_address_form' in request.session:
            billing_address_form = BillingAddressForm(request.session['billing_address_form'])
            billing_address_form.is_valid()
            context['billing_address_form'] = billing_address_form
        if 'credit_card_form' in request.session:
            context['credit_card_form'] = CreditCardForm(request.session['credit_card_form'])

        context = processor.get_checkout_context(context=context)

        return render(request, self.template_name, context)

    def post(self, request, *args, **kwargs):
        # context = super().get_context_data(**kwargs)
        invoice = get_purchase_invoice(request.user, get_site_from_request(request))
        if not invoice.order_items.count() or invoice.status == InvoiceStatus.CART:
            messages.info(request, _("Cart changed while in checkout process"))
            return redirect('vendor:cart')

        processor = get_site_payment_processor(invoice.site)(invoice.site, invoice)

        processor.set_billing_address_form_data(request.session.get('billing_address_form'), BillingAddressForm)
        processor.set_payment_info_form_data(request.session.get('credit_card_form'), CreditCardForm)

        processor.authorize_payment()

        if processor.transaction_succeeded:
            return redirect('vendor:purchase-summary', uuid=invoice.uuid)
        else:
            logger.warning(f"Payment gateway did not authorize payment {processor.transaction_info}")
            # TODO: Make message configurable for the site in the settings
            messages.info(self.request, _("The payment gateway did not authorize payment."))
            return redirect('vendor:checkout-account')


class PaymentSummaryView(LoginRequiredMixin, DetailView):
    model = Invoice
    template_name = 'vendor/payment_summary.html'
    slug_field = 'uuid'
    slug_url_kwarg = 'uuid'

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data()
        context['payment'] = self.object.payments.filter(success=True).first()
        return context


class OrderHistoryListView(LoginRequiredMixin, ListView):
    '''
    List of all the invoices generated by the current user on the current site.
    '''
    model = Invoice
    # TODO: filter to only include the current user's order history

    def get_queryset(self):
        try:
            # The profile and user are site specific so this should only return what's on the site for that user excluding the cart
            return self.request.user.customer_profile.get(site=get_site_from_request(self.request)).invoices.filter(status__gt=InvoiceStatus.CART)
        except ObjectDoesNotExist:         # Catch the actual error for the exception
            return []   # Return empty list if there is no customer_profile


class OrderHistoryDetailView(LoginRequiredMixin, DetailView):
    '''
    Details of an invoice generated by the current user on the current site.
    '''
    template_name = "vendor/invoice_history_detail.html"
    model = Invoice
    slug_field = 'uuid'
    slug_url_kwarg = 'uuid'


class ReceiptListView(LoginRequiredMixin, ListView):
    model = Receipt
    template_name = 'vendor/purchase_list.html'

    def get_queryset(self):
        return self.request.user.customer_profile.get(site=get_site_from_request(self.request)).receipts.all()


class ReceiptDetailView(LoginRequiredMixin, DetailView):
    model = Receipt
    template_name = 'vendor/purchase_detail.html'
    slug_field = 'uuid'
    slug_url_kwarg = 'uuid'

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(**kwargs)

        context['payment'] = self.object.order_item.invoice.payments.get(success=True, transaction=self.object.transaction)

        return context


class SubscriptionsListView(LoginRequiredMixin, ListView):
    model = Receipt
    template_name = 'vendor/purchase_list.html'

    def get_queryset(self):
        try:
            receipts = self.request.user.customer_profile.get(site=get_site_from_request(self.request)).receipts.filter(auto_renew=True)
        except ObjectDoesNotExist:
            raise Http404(_("Not Found"))
        
        subscriptions = [ receipt for receipt in receipts.all() if receipt.order_item.offer.terms > TermType.PERPETUAL and receipt.order_item.offer.terms < TermType.ONE_TIME_USE ]
        
        return subscriptions


class SubscriptionUpdatePaymentView(LoginRequiredMixin, FormView):
    form_class = CreditCardForm()
    success_url = reverse_lazy('vendor:customer-subscriptions')

    def post(self, request, *args, **kwargs):
        subscription = Subscription.objects.get(uuid=self.kwargs["uuid"])
        payment_form = CreditCardForm(request.POST)

        if not payment_form.is_valid():
            messages.info(request, _("Invalid Card"))
            return redirect(request.META.get('HTTP_REFERER', self.success_url))

        processor = get_site_payment_processor(subscription.profile.site)(subscription.profile.site)
        processor.payment_info = payment_form
        processor.subscription_update_payment(subscription)

        if not processor.transaction_succeeded:
            messages.info(request, _(f"Payment gateway error: {processor.transaction_info.get('errors', '')}"))
            return redirect(request.META.get('HTTP_REFERER', self.success_url))

        messages.info(request, _("Success: Payment Updated"))
        return redirect(request.META.get('HTTP_REFERER', self.success_url))


class AddressUpdateView(LoginRequiredMixin, FormView):
    form_class = BillingAddressForm

    def post(self, request, *args, **kwargs):
        address_form = AddressForm(request.POST)
        address = Address.objects.get(uuid=self.kwargs['uuid'])

        if not address_form.is_valid():
            messages.info(request, _(f"Failed: {address_form.errors}"))
        else:
            messages.info(request, _("Success: Address Updated"))
            update_address = address_form.save(commit=False)
            address.first_name = update_address.first_name
            address.last_name = update_address.last_name
            address.address_1 = update_address.address_1
            address.address_2 = update_address.address_2
            address.locality = update_address.locality
            address.state = update_address.state
            address.country = update_address.country
            address.postal_code = update_address.postal_code
            address.save()

        return redirect(request.META.get('HTTP_REFERER', self.success_url))


class ShippingAddressUpdateView(LoginRequiredMixin, UpdateView):
    model = Address
    fields = '__all__'
    template_name_suffix = '_update_form'
    template_name = 'vendor/address_detail.html'
    success_url = reverse_lazy('vendor:products')

    def get_success_url(self):
        messages.info(self.request, _("Shipping Address Updated"))
        return self.request.META.get('HTTP_REFERER', self.success_url)



# Generated by Django 3.2.14 on 2022-07-19 09:44

from django.db import migrations, models
import django.db.models.manager


class Migration(migrations.Migration):

    dependencies = [
        ('vendor', '0037_auto_20220609_1644'),
    ]

    operations = [
        migrations.AlterModelManagers(
            name='payment',
            managers=[
                ('reports', django.db.models.manager.Manager()),
            ],
        ),
        migrations.AlterModelManagers(
            name='subscription',
            managers=[
                ('reports', django.db.models.manager.Manager()),
            ],
        ),
        migrations.AddField(
            model_name='payment',
            name='submitted_date',
            field=models.DateTimeField(blank=True, default=None, null=True, verbose_name='Payment Date'),
        ),
    ]



# Generated by Django 3.1.3 on 2021-01-07 23:10

from django.db import migrations

import uuid


def generate_address_uuid(apps, schema_editor):
    AddressModel = apps.get_model('vendor', 'Address')
    for address in AddressModel.objects.all():
        address.uuid = uuid.uuid4()
        address.save(update_fields=['uuid'])


def generate_customer_profile_uuid(apps, schema_editor):
    CustomerProfileModel = apps.get_model('vendor', 'CustomerProfile')
    for customer_profile in CustomerProfileModel.objects.all():
        customer_profile.uuid = uuid.uuid4()
        customer_profile.save(update_fields=['uuid'])


class Migration(migrations.Migration):

    dependencies = [
        ('vendor', '0018_add_uuid_address_profile'),
    ]

    operations = [
        migrations.RunPython(generate_address_uuid, reverse_code=migrations.RunPython.noop),
        migrations.RunPython(generate_customer_profile_uuid, reverse_code=migrations.RunPython.noop),
    ]
